import pytest

from critical_es_value import ttest


@pytest.mark.parametrize(
    "correction, n1, n2, expected, match",
    [
        (True, 10, 10, True, None),
        (True, 10, 15, True, None),
        (False, 10, 10, False, None),
        (False, 10, 15, False, None),
        ("auto", 10, 10, False, None),
        ("auto", 10, 15, True, None),
        ("auto", 15, 10, True, None),
        ("invalid", 10, 10, None, r"correction must be one of True, False, or 'auto'"),
        ("invalid", 10, 15, None, r"correction must be one of True, False, or 'auto'"),
    ],
)
def test_determine_welch_correction(correction, n1, n2, expected, match):
    if match:
        with pytest.raises(ValueError, match=match):
            ttest.determine_welch_correction(correction, n1=n1, n2=n2)
    else:
        result = ttest.determine_welch_correction(correction, n1=n1, n2=n2)
        assert result == expected


@pytest.mark.parametrize(
    "dataset, alternative, confidence, expected",
    [
        (
            "test_dataset1",
            "two-sided",
            0.95,
            {
                "d": 0.7479307,
                "d_critical": 0.3734061,
                "b_critical": 0.3609507,
                "g": 0.728391,
                "g_critical": 0.3636509,
            },
        ),
        (
            "test_dataset2",
            "two-sided",
            0.95,
            {
                "d": 1.986911,
                "d_critical": 0.2841969,
                "b_critical": 0.3661223,
                "g": 1.956317,
                "g_critical": 0.2798208,
            },
        ),
    ],
)
def test_critical_for_one_sample_ttest(
    test_dataset1, test_dataset2, dataset, alternative, confidence, expected
):
    data = test_dataset1 if dataset == "test_dataset1" else test_dataset2
    result = ttest.critical_for_one_sample_ttest(
        x=data["x"],
        alternative=alternative,
        confidence=confidence,
    ).iloc[0]

    for key, value in expected.items():
        assert result[key] == pytest.approx(value), key


@pytest.mark.parametrize(
    "dataset, paired, alternative, correction, confidence, expected",
    [
        (
            "test_dataset1",
            False,
            "two-sided",
            True,
            0.95,
            {
                "d": 0.5072601,
                "d_critical": 0.5168963,
                "b_critical": 0.5183234,
                "g": 0.5006344,
                "g_critical": 0.5101447,
            },
        ),
        (
            "test_dataset1",
            False,
            "two-sided",
            False,
            0.95,
            {
                "d": 0.5072601,
                "d_critical": 0.5168412,
                "b_critical": 0.5182682,
                "g": 0.5006675,
                "g_critical": 0.5101241,
            },
        ),
        (
            "test_dataset1",
            False,
            "less",
            False,
            0.95,
            {
                "d": 0.5072601,
                "d_critical": -0.4315931,
                "b_critical": -0.4327847,
                "g": 0.5006675,
                "g_critical": -0.4259839,
            },
        ),
        (
            "test_dataset1",
            True,
            "two-sided",
            False,
            0.95,
            {
                "d": 0.5074131,
                "d_critical": 0.6060484,
                "b_critical": 0.6075383,
                "g": 0.494157,
                "g_critical": 0.5902154,
                "dz": 0.3126338,
                "dz_critical": 0.3734061,
                "gz": 0.3044662,
                "gz_critical": 0.3636509,
            },
        ),
        (
            "test_dataset1",
            False,
            "two-sided",
            True,
            0.95,
            {
                "d": 0.5072601,
                "d_critical": 0.5168963,
                "b_critical": 0.5183234,
                "g": 0.5006344,
                "g_critical": 0.5101447,
            },
        ),
        (
            "test_dataset1",
            True,
            "greater",
            False,
            0.90,
            {
                "d": 0.5074131,
                "d_critical": 0.3886078,
                "b_critical": 0.3895632,
                "g": 0.494157,
                "g_critical": 0.3784554,
                "dz": 0.3126338,
                "dz_critical": 0.2394339,
                "gz": 0.3044662,
                "gz_critical": 0.2331787,
            },
        ),
        (
            "test_dataset1",
            True,
            "two-sided",
            False,
            0.99,
            {
                "d": 0.5074131,
                "d_critical": 0.8167803,
                "b_critical": 0.8187883,
                "g": 0.494157,
                "g_critical": 0.7954419,
                "dz": 0.3126338,
                "dz_critical": 0.5032449,
                "gz": 0.3044662,
                "gz_critical": 0.4900976,
            },
        ),
        (
            "test_dataset2",
            False,
            "two-sided",
            True,
            0.95,
            {
                "d": -0.0993843,
                "d_critical": 0.3975426,
                "b_critical": 0.4389254,
                "g": -0.0985214,
                "g_critical": 0.3940911,
            },
        ),
    ],
)
def test_critical_for_two_sample_ttest(
    test_dataset1,
    test_dataset2,
    dataset,
    paired,
    alternative,
    correction,
    confidence,
    expected,
):
    data = test_dataset1 if dataset == "test_dataset1" else test_dataset2
    print(data)
    result = ttest.critical_for_two_sample_ttest(
        x=data["x"],
        y=data["y"],
        paired=paired,
        correction=correction,
        alternative=alternative,
        confidence=confidence,
    ).iloc[0]
    print(result)

    for key, value in expected.items():
        assert result[key] == pytest.approx(value), key



import pytest

from critical_es_value import utils


@pytest.mark.parametrize(
    "confidence, alternative, expected, match",
    [
        (0.90, "less", 0.10, None),
        (0.95, "greater", 0.05, None),
        (0.90, "two-sided", 0.05, None),
        (0.95, "two-sided", 0.025, None),
        (0, "one-sided", 0, r"confidence must be in \(0, 1\)"),
        (1, "one-sided", 0, r"confidence must be in \(0, 1\)"),
        (
            0.95,
            "invalid",
            0,
            r"alternative must be one of 'two-sided', 'greater', or 'less'",
        ),
    ],
)
def test_get_alpha(confidence, alternative, expected, match):
    if match:
        with pytest.raises(ValueError, match=match):
            utils.get_alpha(confidence, alternative)
    else:
        alpha = utils.get_alpha(confidence, alternative)
        assert alpha == pytest.approx(expected)


@pytest.mark.parametrize(
    "dof, expected, match",
    [
        (-1, 0, "dof must be greater than 1."),
        (0, 0, "dof must be greater than 1."),
        (1, 0, "dof must be greater than 1."),
        (2, 0.56418958354, None),
        (5, 0.84074868245, None),
        (7, 0.888202907672, None),
        (10, 0.92274560805, None),
        (20, 0.96194453374, None),
        (30, 0.97475437821, None),
    ],
)
def test_get_bias_correction_factor_J(dof, expected, match):
    if match:
        with pytest.raises(ValueError, match=match):
            utils.get_bias_correction_factor_J(dof)
    else:
        assert utils.get_bias_correction_factor_J(dof) == pytest.approx(expected)



from typing import Optional, Union

import numpy as np
import pandas as pd
import pingouin
from numpy.typing import ArrayLike
from scipy import stats

from critical_es_value import utils


def determine_welch_correction(correction: Union[bool, str], n1: int, n2: int) -> bool:
    """Determine whether to apply Welch's correction for unequal variances.

    Args:
        correction (bool or str): If True, always apply Welch's correction.
                                  If False, never apply Welch's correction.
                                  If "auto", apply Welch's correction only if sample sizes are unequal.
        n1 (int): Sample size of group 1.
        n2 (int): Sample size of group 2.

    Returns:
        bool: Whether to apply Welch's correction.

    Raises:
        ValueError: If `correction` is not one of True, False, or "auto".
    """
    if correction not in (True, False, "auto"):
        raise ValueError("correction must be one of True, False, or 'auto'")
    if correction is True or (correction == "auto" and n1 != n2):
        return True
    return False


def critical_for_one_sample_ttest_from_values(
    t: float,
    n: int,
    dof: int,
    confidence: float = 0.95,
    alternative: str = "two-sided",
    std: Optional[float] = None,
):
    """Calculate critical effect size values for a one-sample t-test given t, sample size and other parameters.

    Args:
        t (float): t-value of the test statistic.
        n (int): Sample size.
        dof (int): Degrees of freedom.
        confidence (float): Confidence level between 0 and 1 (exclusive). Default is 0.95.
        alternative (str): The alternative hypothesis. Either "two-sided", "greater", or "less". Default is "two-sided".
        std (Optional[float]): Standard deviation of the sample. If None, b_critical will not be calculated. Default is None.

    Returns:
        pd.DataFrame: Returns a DataFrame with the following columns:
            - `T`: t-value of the test statistic
            - `dof`: Degrees of freedom
            - `T_critical`: Critical t-value
            - `d`: Cohen's d
            - `d_critical`: Critical value for Cohen's d
            - `g`: Hedges' g
            - `g_critical`: Critical value for Hedges' g
            - `b_critical`: Critical value for the raw mean difference
    """
    alpha = utils.get_alpha(confidence, alternative)

    factor = np.sqrt(1 / n)
    d = t * factor

    tc = np.abs(stats.t.ppf(alpha, dof))
    if alternative == "less":
        tc *= -1

    dc = tc * factor
    j = utils.get_bias_correction_factor_J(dof)

    result = {
        "T": t,
        "dof": dof,
        "T_critical": tc,
        "d": d,
        "d_critical": dc,
        "g": d * j,
        "g_critical": dc * j,
    }
    if std is not None:
        result["b_critical"] = tc * std / np.sqrt(n)

    return pd.DataFrame([result], index=["critical"])


def critical_for_one_sample_ttest(
    x: ArrayLike,
    confidence: float = 0.95,
    alternative: str = "two-sided",
) -> pd.DataFrame:
    """Calculate critical effect size values for a one-sample t-test.

    Args:
        x (ArrayLike): Sample data.
        confidence (float): Confidence level between 0 and 1 (exclusive). Default is 0.95.
        alternative (str): The alternative hypothesis. Either "two-sided", "greater", or "less". Default is "two-sided".

    Returns:
        pd.DataFrame: Returns a DataFrame with the following columns:
            - `T`: t-value of the test statistic
            - `dof`: Degrees of freedom
            - `T_critical`: Critical t-value
            - `d`: Cohen's d
            - `d_critical`: Critical value for Cohen's d
            - `g`: Hedges' g
            - `g_critical`: Critical value for Hedges' g
            - `b_critical`: Critical value for the raw mean difference
    """

    t_test_result = pingouin.ttest(
        x=x,
        y=0,
        paired=False,
        correction=False,
        confidence=confidence,
        alternative=alternative,
    ).iloc[0]

    return critical_for_one_sample_ttest_from_values(
        t=t_test_result["T"],
        n=len(x),
        confidence=confidence,
        alternative=alternative,
        dof=t_test_result.dof,
        std=np.std(x, ddof=1),
    )


def _critical_for_two_sample_ttest_unpaired_from_values(
    t: float,
    n1: int,
    n2: int,
    dof: int,
    std1: Optional[float] = None,
    std2: Optional[float] = None,
    correction: Union[bool, str] = "auto",
    confidence: float = 0.95,
    alternative: str = "two-sided",
) -> pd.DataFrame:
    """Calculate critical effect size values for UNPAIRED two-sample t-test given t, sample sizes and other parameters.

    Args:
        t (float): t-value of the test statistic.
        n1 (int): Sample size of group 1.
        n2 (int): Sample size of group 2.
        dof (int): Degrees of freedom.
        std1 (Optional[float]): Standard deviation of group 1. If None, b_critical will not be calculated. Default is None.
        std2 (Optional[float]): Standard deviation of group 2. If None, b_critical will not be calculated. Default is None.
        correction (bool): Specify whether or not to correct for unequal variances using Welch separate variances
            T-test. If "auto", it will automatically uses Welch T-test when the sample sizes are unequal. Default
            is "auto".
        confidence (float): Confidence level between 0 and 1 (exclusive). Default is 0.95.
        alternative (str): The alternative hypothesis. Either "two-sided", "greater", or "less". Default is "two-sided".

    Returns:
        pd.DataFrame: Returns a DataFrame with the following columns:
           - `T`: t-value of the test statistic
           - `dof`: Degrees of freedom
           - `T_critical`: Critical t-value
           - `d`: Cohen's d
           - `d_critical`: Critical value for Cohen's d
           - `g`: Hedges' g
           - `g_critical`: Critical value for Hedges' g
           - `b_critical`: Critical value for the raw mean difference (if std1 and std2 are provided)
    """

    alpha = utils.get_alpha(confidence, alternative)

    factor = np.sqrt(1 / n1 + 1 / n2)
    d = t * factor

    tc = np.abs(stats.t.ppf(alpha, dof))
    if alternative == "less":
        tc *= -1
    dc = tc * factor

    if std1 is not None and std2 is not None:
        if determine_welch_correction(correction, n1=n1, n2=n2):
            se = np.sqrt((std1**2 / n1) + (std2**2 / n2))
        else:
            se = np.sqrt((std1**2 * (n1 - 1) + std2**2 * (n2 - 1)) / (n1 + n2 - 2))
            se *= factor

    j = utils.get_bias_correction_factor_J(dof)

    result = {
        "T": t,
        "dof": dof,
        "T_critical": tc,
        "d": d,
        "d_critical": dc,
        "g": d * j,
        "g_critical": dc * j,
    }
    if std1 is not None and std2 is not None:
        result["b_critical"] = tc * se

    return pd.DataFrame([result], index=["critical"])


def _critical_for_two_sample_ttest_paired_from_values(
    t: float,
    n: int,
    dof: int,
    r12: float,
    confidence: float = 0.95,
    alternative: str = "two-sided",
    std_diff: Optional[float] = None,
) -> pd.DataFrame:
    """Calculate critical effect size values for a PAIRED two-sample t-test given t, sample sizes and other parameters.

    Args:
        t (float): t-value of the test statistic.
        n (int): Sample sizes of both groups.
        dof (int): Degrees of freedom.
        r12 (float): Pearson correlation between the two groups.
        confidence (float): Confidence level between 0 and 1 (exclusive). Default is 0.95.
        alternative (str): The alternative hypothesis. Either "two-sided", "greater", or "less". Default is "two-sided".
        std_diff (Optional[float]): Standard deviation of the difference scores. If None, b_critical will not be calculated. Default is None.

    Returns:
        pd.DataFrame: Returns a DataFrame with the following columns:
            - `T`: t-value of the test statistic
            - `dof`: Degrees of freedom
            - `T_critical`: Critical t-value
            - `d`: Cohen's d
            - `d_critical`: Critical value for Cohen's d
            - `g`: Hedges' g
            - `g_critical`: Critical value for Hedges' g
            - `dz`: Cohen's dz
            - `dz_critical`: Critical value for Cohen's dz
            - `gz`: Hedges' gz
            - `gz_critical`: Critical value for Hedges' gz
            - `b_critical`: Critical value for the raw mean difference (if std_diff is provided)
    """

    alpha = utils.get_alpha(confidence, alternative)

    factor1 = np.sqrt(1 / n)
    factor2 = np.sqrt(2 * (1 - r12))

    dz = t * factor1
    d = dz * factor2

    tc = np.abs(stats.t.ppf(alpha, dof))
    if alternative == "less":
        tc *= -1

    dzc = tc * factor1
    dc = dzc * factor2

    j = utils.get_bias_correction_factor_J(dof)

    result = {
        "T": t,
        "dof": dof,
        "T_critical": tc,
        "d": d,
        "d_critical": dc,
        "g": d * j,
        "g_critical": dc * j,
        "dz": dz,
        "dz_critical": dzc,
        "gz": dz * j,
        "gz_critical": dzc * j,
    }
    if std_diff is not None:
        result["b_critical"] = tc * std_diff / np.sqrt(n)

    return pd.DataFrame([result], index=["critical"])


def critical_for_two_sample_ttest_from_values(
    t: float,
    n1: int,
    n2: int,
    dof: int,
    std1: Optional[float] = None,
    std2: Optional[float] = None,
    paired: bool = False,
    correction: Union[bool, str] = "auto",
    confidence: float = 0.95,
    alternative: str = "two-sided",
) -> pd.DataFrame:
    """Calculate critical effect size values for a paired or an unpaired two-sample t-test given t, sample sizes and other parameters.

    Args:
        t (float): t-value of the test statistic.
        n1 (int): Sample size of group 1.
        n2 (int): Sample size of group 2.
        dof (int): Degrees of freedom.
        std1 (Optional[float]): Standard deviation of group 1. If None, b_critical will not be calculated. Default is None.
        std2 (Optional[float]): Standard deviation of group 2. If None, b_critical will not be calculated. Default is None.
        paired (bool): Whether the samples are paired. Default is False.
        correction (bool): For unpaired two sample T-tests, specify whether or not to correct for unequal variances
            using Welch separate variances T-test. If "auto", it will automatically uses Welch T-test when the sample
            sizes are unequal. For paired T-tests, this parameter is ignored and no correction is performed. Default
            is "auto".
        confidence (float): Confidence level between 0 and 1 (exclusive). Default is 0.95.
        alternative (str): The alternative hypothesis. Either "two-sided", "greater", or "less". Default is "two-sided".

    Returns:
        pd.DataFrame: Returns a DataFrame with the following columns:
           - `T`: t-value of the test statistic
           - `dof`: Degrees of freedom
           - `T_critical`: Critical t-value
           - `d`: Cohen's d
           - `d_critical`: Critical value for Cohen's d
           - `g`: Hedges' g
           - `g_critical`: Critical value for Hedges' g
           - `dz`: Cohen's dz (only for paired tests)
           - `dz_critical`: Critical value for Cohen's dz (only for paired tests)
           - `gz`: Hedges' gz (only for paired tests)
           - `gz_critical`: Critical value for Hedges' gz (only for paired tests)
           - `b_critical`: Critical value for the raw mean difference
    """
    if paired:
        return _critical_for_two_sample_ttest_paired_from_values(
            t=t,
            n1=n1,
            n2=n2,
            dof=dof,
            confidence=confidence,
            alternative=alternative,
        )

    return _critical_for_two_sample_ttest_unpaired_from_values(
        t=t,
        n1=n1,
        n2=n2,
        dof=dof,
        std1=std1,
        std2=std2,
        confidence=confidence,
        alternative=alternative,
        correction=correction,
    )


def critical_for_two_sample_ttest(
    x: ArrayLike,
    y: ArrayLike,
    paired: bool = False,
    correction: Union[bool, str] = "auto",
    confidence: float = 0.95,
    alternative: str = "two-sided",
) -> pd.DataFrame:
    """Calculate critical effect size values for a paired or an unpaired two-sample t-test.

    Args:
        x (ArrayLike): Sample data for group 1.
        y (ArrayLike): Sample data for group 2.
        paired (bool): Whether the samples are paired. Default is False.
        correction (bool): For unpaired two sample T-tests, specify whether or not to correct for unequal variances
            using Welch separate variances T-test. If "auto", it will automatically uses Welch T-test when the sample
            sizes are unequal. For paired T-tests, this parameter is ignored and no correction is performed. Default
            is "auto".
        confidence (float): Confidence level between 0 and 1 (exclusive). Default is 0.95.
        alternative (str): The alternative hypothesis. Either "two-sided", "greater", or "less". Default is "two-sided".

    Returns:
        pd.DataFrame: Returns a DataFrame with the following columns:
           - `T`: t-value of the test statistic
           - `dof`: Degrees of freedom
           - `T_critical`: Critical t-value
           - `d`: Cohen's d
           - `d_critical`: Critical value for Cohen's d
           - `g`: Hedges' g
           - `g_critical`: Critical value for Hedges' g
           - `b_critical`: Critical value for the raw mean difference
    """
    if paired:
        if len(x) != len(y):
            raise ValueError("For paired tests, x and y must have the same length.")

        t_test_result = pingouin.ttest(
            x=x,
            y=y,
            paired=True,
            correction=False,
            confidence=confidence,
            alternative=alternative,
        ).iloc[0]

        return _critical_for_two_sample_ttest_paired_from_values(
            t=t_test_result["T"],
            n=len(x),
            dof=t_test_result.dof,
            r12=np.corrcoef(x, y)[0, 1],
            confidence=confidence,
            alternative=alternative,
            std_diff=np.std(x - y, ddof=1),
        )

    n1 = len(x)
    n2 = len(y)
    correction = determine_welch_correction(correction, n1=n1, n2=n2)

    t_test_result = pingouin.ttest(
        x=x,
        y=y,
        paired=False,
        correction=correction,
        confidence=confidence,
        alternative=alternative,
    ).iloc[0]

    return _critical_for_two_sample_ttest_unpaired_from_values(
        t=t_test_result["T"],
        n1=n1,
        n2=n2,
        dof=t_test_result.dof,
        std1=np.std(x, ddof=1),
        std2=np.std(y, ddof=1),
        confidence=confidence,
        alternative=alternative,
        correction=correction,
    )
